from collections import defaultdict

def startswith(s1,s2):
  """Check is list s1 starts with list s2"""
  l = len(s2)
  return len(s1) >= l and s1[0:l] == s2

def parse_mt_pair(tp):
  """Parses a pair of Maude meta-terms separated by a comma"""
  t1,t2 = "",""
  stack,i,shift,nobrace = 0,0,1,True
  while True:
    if tp[i] == '`':
      t1 += tp[i]
      i += 1
    elif tp[i] == '[':
      stack += 1
      nobrace = False
    elif tp[i] == ']' or (tp[i] == ',' and nobrace):
      stack -= 1
      if stack == 0 or nobrace:
        t1 += tp[i]
        break
    t1 += tp[i]
    i += 1
  #
  if not nobrace: shift += 1
  t2 = tp[i+shift:] 
  return (t1,t2)

def findnth(haystack, needle, n):
  """Finds the nth occurence of str needle in str haystack"""
  parts = haystack.split(needle, n+1)
  if len(parts)<=n+1:
    return -1
  return len(haystack)-len(parts[-1])-len(needle)

def getexpr(lines,i):
  """Parses the expression being rewritten (i.e. the experiment being run"""
  if not startswith(lines[i],"rewrite in RUN : "):
    raise ValueError("Unexpected string: line "+str(i))
  expr = lines[i].strip()
  #
  while i + 1 < len(lines):
    i += 1
    if not startswith(lines[i],"rewrites: "):
      expr += lines[i].strip()
      continue
    break
  # 
  typ,val,num = None,None,None
  try:
    # get the type: ms or std
    pos1 = expr.index("-")+1
    pos2 = expr[pos1:].index("-")
    typ = expr[pos1:pos1+pos2]
    # get the arguments (between the parens)
    pos1 = findnth(expr,"(",1)+1
    if pos1 == -1: raise ValueError()
    pos2 = expr.index(")")
    val = expr[pos1:pos2]
    # get the depth of this experiment
    pos1 = expr.index(",")+1
    pos2 = expr.index("&")
    num = int(expr[pos1:pos2].strip())
  except ValueError:
    raise ValueError("Unexpected string: line "+str(i))
  # extract the two halves of the argument
  arg1,arg2 = parse_mt_pair(val)
  return (typ,num,arg1,arg2,i)

def get_rewrites(lines,i):
  """Parses the # of rewrites and time used to generate an answer"""
  if not startswith(lines[i],"rewrites: "):
    raise ValueError("Unexpected string: line "+str(i))
  rew,time = None,None
  rewdata = ""
  while i < len(lines):
    if not startswith(lines[i],"result "):
      rewdata += lines[i].strip()
      i += 1
      continue
    break
  # 
  try:
    pos1 = 10
    pos2 = rewdata.index("in") - 1
    rew = int(rewdata[pos1:pos2].strip())
    pos3 = pos2 + 3
    pos4 = rewdata.index("ms")
    time = int(rewdata[pos3:pos4])
  except ValueError:
    raise ValueError("Unexpected string: line "+str(i))
  return (rew,time,i)

def get_result(lines,i):
  """Parses the result of the experiment: not fully implemented yet"""
  if not startswith(lines[i],"result "):
    raise ValueError("Unexpected string")
  try:
    start = lines[i].index(": ") + 2
  except ValueError:
    raise ValueError("Unexpected string")
  return lines[i]

def ignore_line(lines,i):
  """Ignores any line which does not start with our expected values"""
  if startswith(lines[i],"rewrites: ") or    \
     startswith(lines[i],"rewrite in RUN : "):
    return False
  return True

def parse_file(fn,data):
  """Parses an experimental results file"""
  with open(fn,"r") as f:
    lines = f.readlines()
    ln,end = 0,len(lines)
    curr = None
    count = 0
    # Either:
    # 1) Ignore the line
    # 2) Process it to get the expression
    # 3) Process it to get the rewrites and time used
    while ln < end:
      if ignore_line(lines,ln):
        ln += 1
        continue
      elif curr == None:
        curr = getexpr(lines,ln)
        ln = curr[4]
      else:
        rt = get_rewrites(lines,ln)
        data[curr[2]+curr[3]].append((curr[0],curr[1],curr[2],curr[3],rt[0],rt[1]))
        curr = None
        count += 1
        ln += 1
    # Ignore any unfnished entries
    # Print statement below useful for debugging purposes
    # print("# ENTRIES: "+str(count))
    return data

def getExprRewTimeData(osfile,msfile):
  # get the results of both experiments 
  datax = parse_file(osfile,defaultdict(list))
  dataxy = parse_file(msfile,datax)
  # prune the result sets so that only the experiments that finished
  # in both categories are recorded
  sharedrewx = list()
  sharedrewy = list()
  sharedtimex = list()
  sharedtimey = list()
  for vs in dataxy.values():
    if len(vs) > 2 or len(vs) == 0:
      raise ValueError("Malformed dict")
    elif len(vs) < 2:
      continue
    else:
      sharedrewx.append(vs[0][4])
      sharedrewy.append(vs[1][4])
      sharedtimex.append(vs[0][5])
      sharedtimey.append(vs[1][5])
  # return all of the collected data
  return (sharedrewx,sharedrewy,sharedtimex,sharedtimey)

def prune(datax,datay,nx,ny,p_flag):
  # prune unwanted data points even further which distort the graph
  topdatay = sorted(datay)[-1*ny:]
  topdatax = sorted(datax)[-1*nx:]
  px,py = list(),list()
  # iterate through all points in the graph
  for v1,v2 in zip(datax,datay):
    # if either dimension is too far, prune the point
    if v2 in topdatay:
      if p_flag: print(str(v1)+" "+str(v2))
      continue
    if v1 in topdatax:
      if p_flag: print(str(v1)+" "+str(v2))
      continue
    px.append(v1)
    py.append(v2)
  # return the pruned x and y sets
  return (px,py)
